import logging
import operator
import warnings
from typing import List, Literal, Optional, Tuple

import numpy as np
import pandas as pd
import xgboost as xgb

from bluecast.preprocessing.custom import CustomPreprocessing


class BoostaRootaWrapper(CustomPreprocessing):
    def __init__(
        self, class_problem: Literal["binary", "multiclass", "regression"], random_state
    ):
        super().__init__()
        self.class_problem = class_problem
        self.selected_features: List[Optional[str]] = []
        self.random_state = random_state

        if self.class_problem not in ["binary", "multiclass", "regression"]:
            raise ValueError(
                "Class_problem must be one of 'binary', 'multiclass', or 'regression'."
            )

    def fit_transform(
        self, df: pd.DataFrame, targets: pd.Series
    ) -> Tuple[pd.DataFrame, Optional[pd.Series]]:
        if self.class_problem == "binary":
            model = xgb.XGBClassifier(
                tree_method="approx",
                max_bin=255,
                n_estimators=100,
                random_state=self.random_state,
                importance_type="total_cover",
                max_depth=10,
            )
            br = BoostARoota(clf=model)
        elif self.class_problem == "multiclass":
            br = BoostARoota(metric="mlogloss")
        else:
            model = xgb.XGBRegressor(
                tree_method="approx",
                max_bin=255,
                n_estimators=100,
                random_state=self.random_state,
                importance_type="total_cover",
                max_depth=10,
            )
            br = BoostARoota(clf=model)

        logging.info("Start feature selection as defined in FeatureSelectionConfig.")
        br.fit(df, targets)
        self.selected_features = br.keep_vars_
        logging.info(
            f"{len(self.selected_features)} from {len(df.columns)} features selected."
        )
        logging.info("Selected features are {self.selected_features}).")
        return df.loc[:, self.selected_features], targets

    def transform(
        self,
        df: pd.DataFrame,
        target: Optional[pd.Series] = None,
        predicton_mode: bool = False,
    ) -> Tuple[pd.DataFrame, Optional[pd.Series]]:
        logging.info("Apply feature selection.")
        df = df.loc[:, self.selected_features]
        return df, target


class BoostARoota(object):

    def __init__(
        self,
        metric=None,
        clf=None,
        cutoff=200,
        iters=10,
        max_rounds=100,
        delta=0.1,
        silent=True,
    ):
        self.metric = metric
        self.clf = clf
        self.cutoff = cutoff
        self.iters = iters
        self.max_rounds = max_rounds
        self.delta = delta
        self.silent = silent
        self.keep_vars_ = None

        # Throw errors if the inputted parameters don't meet the necessary criteria
        if (metric is None) and (clf is None):
            raise ValueError("you must enter one of metric or clf as arguments")
        if cutoff <= 0:
            raise ValueError(
                "cutoff should be greater than 0. You entered" + str(cutoff)
            )
        if iters <= 0:
            raise ValueError("iters should be greater than 0. You entered" + str(iters))
        if (delta <= 0) | (delta > 1):
            raise ValueError("delta should be between 0 and 1, was " + str(delta))

        # Issue warnings for parameters to still let it run
        if (metric is not None) and (clf is not None):
            warnings.warn(
                "You entered values for metric and clf, defaulting to clf and ignoring metric",
                stacklevel=2,
            )
        if delta < 0.02:
            warnings.warn(
                "WARNING: Setting a delta below 0.02 may not converge on a solution.",
                stacklevel=2,
            )
        if max_rounds < 1:
            warnings.warn(
                "WARNING: Setting max_rounds below 1 will automatically be set to 1.",
                stacklevel=2,
            )

    def fit(self, x, y):
        self.keep_vars_ = _BoostARoota(
            x,
            y,
            metric=self.metric,
            clf=self.clf,
            cutoff=self.cutoff,
            iters=self.iters,
            max_rounds=self.max_rounds,
            delta=self.delta,
            silent=self.silent,
        )
        return self

    def transform(self, x):
        if self.keep_vars_ is None:
            raise ValueError("You need to fit the model first")
        return x[self.keep_vars_]

    def fit_transform(self, x, y):
        self.fit(x, y)
        return self.transform(x)


def _create_shadow(x_train):
    """
    Take all X variables, creating copies and randomly shuffling them
    :param x_train: the dataframe to create shadow features on
    :return: dataframe 2x width and the names of the shadows for removing later
    """
    x_shadow = x_train.copy()
    for c in x_shadow.columns:
        np.random.shuffle(x_shadow[c].values)
    # rename the shadow
    shadow_names = ["ShadowVar" + str(i + 1) for i in range(x_train.shape[1])]
    x_shadow.columns = shadow_names
    # Combine to make one new dataframe
    new_x = pd.concat([x_train, x_shadow], axis=1)
    return new_x, shadow_names


def _reduce_vars_xgb(x, y, metric, this_round, cutoff, n_iterations, delta, silent):
    """
    Function to run through each
    :param x: Input dataframe - X
    :param y: Target variable
    :param metric: Metric to optimize in XGBoost
    :param this_round: Round so it can be printed to screen
    :return: tuple - stopping criteria and the variables to keep
    """
    # Set up the parameters for running the model in XGBoost - split is on multi log loss
    if metric == "mlogloss":
        param = {
            "objective": "multi:softmax",
            "eval_metric": "mlogloss",
            "num_class": len(np.unique(y)),
            "silent": 1,
        }
    else:
        param = {"eval_metric": metric, "silent": 1}
    for i in range(1, n_iterations + 1):
        # Create the shadow variables and run the model to obtain importances
        new_x, shadow_names = _create_shadow(x)
        dtrain = xgb.DMatrix(new_x, label=y, enable_categorical=True)
        bst = xgb.train(param, dtrain, verbose_eval=False)
        if i == 1:
            df = pd.DataFrame({"feature": new_x.columns})
            pass

        importance = bst.get_fscore()
        importance = sorted(importance.items(), key=operator.itemgetter(1))
        df2 = pd.DataFrame(importance, columns=["feature", "fscore" + str(i)])
        df2["fscore" + str(i)] = df2["fscore" + str(i)] / df2["fscore" + str(i)].sum()
        df = pd.merge(df, df2, on="feature", how="outer")
        if not silent:
            print("Round: ", this_round, " iteration: ", i)

    df_numeric = df.select_dtypes(include=[np.number])  # Only select numeric columns
    df["Mean"] = df_numeric.mean(axis=1)
    # Split them back out
    real_vars = df[~df["feature"].isin(shadow_names)]
    shadow_vars = df[df["feature"].isin(shadow_names)]

    # Get mean value from the shadows
    mean_shadow = shadow_vars["Mean"].mean() / cutoff
    real_vars = real_vars[(real_vars.Mean > mean_shadow)]

    # Check for the stopping criteria
    # Basically looking to make sure we are removing at least 10% of the variables, or we should stop
    if (len(real_vars["feature"]) / len(x.columns)) > (1 - delta):
        criteria = True
    else:
        criteria = False

    return criteria, real_vars["feature"]


def _reduce_vars_sklearn(x, y, clf, this_round, cutoff, n_iterations, delta, silent):
    """
    Function to run through each
    :param x: Input dataframe - X
    :param y: Target variable
    :param clf: the fully specified classifier passed in by user
    :param this_round: Round so it can be printed to screen
    :return: tuple - stopping criteria and the variables to keep
    """
    # Set up the parameters for running the model in XGBoost - split is on multi log loss

    for i in range(1, n_iterations + 1):
        # Create the shadow variables and run the model to obtain importances
        new_x, shadow_names = _create_shadow(x)
        clf = clf.fit(new_x, np.ravel(y))

        if i == 1:
            df = pd.DataFrame({"feature": new_x.columns})
            df2 = df.copy()
            pass

        try:
            importance = clf.feature_importances_
            df2["fscore" + str(i)] = importance
        except ValueError:
            print(
                "this clf doesn't have the feature_importances_ method.  Only Sklearn tree based methods allowed"
            )

        # importance = sorted(importance.items(), key=operator.itemgetter(1))

        # df2 = pd.DataFrame(importance, columns=['feature', 'fscore'+str(i)])
        df2["fscore" + str(i)] = df2["fscore" + str(i)] / df2["fscore" + str(i)].sum()
        df = pd.merge(df, df2, on="feature", how="outer", suffixes=("", "_y"))
        df.drop(df.filter(regex="_y$").columns, axis=1, inplace=True)
        if not silent:
            print("Round: ", this_round, " iteration: ", i)

    df["Mean"] = df.drop("feature", axis=1).mean(axis=1)
    # Split them back out
    real_vars = df[~df["feature"].isin(shadow_names)]
    shadow_vars = df[df["feature"].isin(shadow_names)]

    # Get mean value from the shadows
    mean_shadow = shadow_vars["Mean"].mean() / cutoff
    real_vars = real_vars[(real_vars.Mean > mean_shadow)]

    # Check for the stopping criteria
    # Basically looking to make sure we are removing at least 10% of the variables, or we should stop
    if (len(real_vars["feature"]) / len(x.columns)) > (1 - delta):
        criteria = True
    else:
        criteria = False

    return criteria, real_vars["feature"]


# Main function exposed to run the algorithm
def _BoostARoota(x, y, metric, clf, cutoff, iters, max_rounds, delta, silent):
    """
    Function loops through, waiting for the stopping criteria to change
    :param x: X dataframe One Hot Encoded
    :param y: Labels for the target variable
    :param metric: The metric to optimize in XGBoost
    :return: names of the variables to keep
    """

    new_x = x.copy()
    # Run through loop until "crit" changes
    i = 0
    while True:
        # Inside this loop we reduce the dataset on each iteration exiting with keep_vars
        i += 1
        if clf is None:
            crit, keep_vars = _reduce_vars_xgb(
                new_x,
                y,
                metric=metric,
                this_round=i,
                cutoff=cutoff,
                n_iterations=iters,
                delta=delta,
                silent=silent,
            )
        else:
            crit, keep_vars = _reduce_vars_sklearn(
                new_x,
                y,
                clf=clf,
                this_round=i,
                cutoff=cutoff,
                n_iterations=iters,
                delta=delta,
                silent=silent,
            )

        if crit | (i >= max_rounds):
            break  # exit and use keep_vars as final variables
        else:
            new_x = new_x[keep_vars].copy()
    if not silent:
        print("BoostARoota ran successfully! Algorithm went through ", i, " rounds.")
    return keep_vars



import numpy as np
import pandas as pd

from bluecast.blueprints.cast import BlueCast
from bluecast.config.training_config import TrainingConfig
from bluecast.conformal_prediction.nonconformity_measures import (
    brier_score,
    margin_nonconformity_measure,
)


def test_bluecast_fit_binary_without_custom_model():
    # Create an instance of the BlueCast class with the custom model
    train_config = TrainingConfig()
    train_config.calculate_shap_values = False

    bluecast = BlueCast(
        class_problem="binary",
        conf_training=train_config,
    )

    # Create some sample data for testing
    x_train = pd.DataFrame(
        {
            "feature1": [i for i in range(20)],
            "feature2": [i for i in range(20)],
            "feature3": [i for i in range(20)],
            "feature4": [i for i in range(20)],
            "feature5": [i for i in range(20)],
            "feature6": [i + 4 for i in range(20)],
        }
    )
    y_train = pd.Series([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1])
    y_train = y_train.replace({0: "zero", 1: "one"})

    x_test = pd.DataFrame(
        {
            "feature1": [i for i in range(20)],
            "feature2": [i for i in range(20)],
            "feature3": [i for i in range(20)],
            "feature4": [i for i in range(20)],
            "feature5": [i for i in range(20)],
            "feature6": [i + 4 for i in range(20)],
        }
    )
    y_test = pd.Series([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1])
    y_test = y_test.replace({0: "zero", 1: "one"})

    x_calibration = pd.DataFrame(
        {
            "feature1": [i for i in range(20)],
            "feature2": [i for i in range(20)],
            "feature3": [i for i in range(20)],
            "feature4": [i for i in range(20)],
            "feature5": [i for i in range(20)],
            "feature6": [i + 4 for i in range(20)],
        }
    )
    y_calibration = pd.Series(
        [0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1]
    )
    y_calibration = y_calibration.replace({0: "zero", 1: "one"})

    x_train["target"] = y_train

    train_all = pd.concat([x_train, x_test], axis=0)
    train_all["target"] = pd.concat([y_train, y_test], axis=0)

    # Fit the BlueCast model using the custom model
    bluecast.fit(train_all, "target")

    # Predict on the test data using the custom model
    predicted_probas, predicted_classes = bluecast.predict(x_test)
    print("++++++++")
    print("BINARY PREDICTED PROBAS")
    print(predicted_probas)

    # Assert the expected results
    print(predicted_probas.shape)
    print(type(predicted_probas))
    # assert predicted_probas.shape[1] == 1

    probas = bluecast.predict_proba(x_test)
    print("--------")
    print("BINARY PREDICTED PROBAS from predict_proba")
    print(probas)

    # test conformal prediction
    bluecast.calibrate(x_calibration, y_calibration)
    pred_intervals = bluecast.predict_p_values(x_test)
    pred_sets = bluecast.predict_sets(x_test)
    assert isinstance(pred_intervals, np.ndarray)
    assert isinstance(pred_sets, pd.DataFrame)
    print(f"BINARY pred_intervals: {pred_intervals}")
    print(f"BINARY pred_sets: {pred_sets}")

    # test passing brier score
    bluecast.calibrate(
        x_calibration, y_calibration, **{"nonconformity_measure_scorer": brier_score}
    )
    pred_intervals = bluecast.predict_p_values(x_test)
    pred_sets = bluecast.predict_sets(x_test)
    assert isinstance(pred_intervals, np.ndarray)
    assert isinstance(pred_sets, pd.DataFrame)
    print(f"BINARY pred_intervals: {pred_intervals}")
    print(f"BINARY pred_sets: {pred_sets}")

    # test passing MNM score
    bluecast.calibrate(
        x_calibration,
        y_calibration,
        **{"nonconformity_measure_scorer": margin_nonconformity_measure},
    )
    pred_intervals = bluecast.predict_p_values(x_test)
    pred_sets = bluecast.predict_sets(x_test)
    assert isinstance(pred_intervals, np.ndarray)
    assert isinstance(pred_sets, pd.DataFrame)
    print(f"BINARY pred_intervals: {pred_intervals}")
    print(f"BINARY pred_sets: {pred_sets}")



from unittest.mock import patch

import xgboost as xgb

from bluecast.general_utils.general_utils import check_gpu_support


def test_check_gpu_support_cpu_fallback():
    """Test that the function falls back to CPU when no GPU is detected."""

    # Mock xgb.train to always raise an XGBoostError, simulating no GPU support
    with patch("xgboost.train", side_effect=xgb.core.XGBoostError("No GPU support")):
        params = check_gpu_support()
        assert params == {"tree_method": "hist", "device": "cpu"}


def test_check_gpu_support_gpu_available():
    """Test that the function detects GPU support correctly."""

    # Mock xgb.train to work correctly, simulating GPU support
    with patch("xgboost.train") as mock_train:
        params = check_gpu_support()
        # Verify that GPU parameters are returned
        assert params in [
            {"device": "cuda", "tree_method": "gpu_hist"},
            {"device": "cuda"},
            {"tree_method": "gpu_hist"},
        ]

        # Ensure that xgb.train was called at least once
        mock_train.assert_called()


def test_check_gpu_support_gpu_warning():
    """Test that the function falls back to CPU if GPU-related warnings are captured."""

    # Mock xgb.train to raise a warning with 'GPU' in the message, simulating GPU issues
    with patch("xgboost.train"):
        with patch("warnings.warn") as mock_warn:
            mock_warn.side_effect = lambda *args, **kwargs: mock_warn.message
            mock_warn.message = "GPU-related warning"

            params = check_gpu_support()
            assert params == {"tree_method": "gpu_hist", "device": "cuda"}



from typing import Tuple
from unittest.mock import patch

import pandas as pd
import pytest

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostRegressionFinalParamConfig,
    XgboostTuneParamsRegressionConfig,
)
from bluecast.ml_modelling.xgboost_regression import XgboostModelRegression
from bluecast.tests.make_data.create_data import create_synthetic_dataframe_regression


@pytest.fixture
def default_model():
    return XgboostModelRegression(class_problem="regression")


@pytest.fixture
def synthetic_train_test_data_xgboost() -> (
    Tuple[pd.DataFrame, pd.Series, pd.DataFrame, pd.Series]
):
    df_train = create_synthetic_dataframe_regression(2000, random_state=20)
    df_val = create_synthetic_dataframe_regression(2000, random_state=200)
    y_train = df_train.pop("target")
    y_test = df_val.pop("target")
    return df_train, y_train, df_val, y_test


def test_check_load_confs_defaults(default_model):
    with patch("bluecast.ml_modelling.xgboost_regression.logging") as mock_logger:
        default_model.check_load_confs()

        # Check if the defaults are loaded
        assert isinstance(default_model.conf_training, TrainingConfig)
        assert isinstance(default_model.conf_xgboost, XgboostTuneParamsRegressionConfig)
        assert isinstance(
            default_model.conf_params_xgboost, XgboostRegressionFinalParamConfig
        )

        # Ensure logger was called with default config loading messages
        assert mock_logger.call_count == 0


def test_check_load_confs_partial():
    conf_training = TrainingConfig()
    model = XgboostModelRegression(
        class_problem="regression", conf_training=conf_training
    )

    with patch("bluecast.ml_modelling.xgboost_regression.logging") as mock_logger:
        model.check_load_confs()

        # Check if the provided config is used and defaults are loaded for the rest
        assert model.conf_training is conf_training
        assert isinstance(model.conf_xgboost, XgboostTuneParamsRegressionConfig)
        assert isinstance(model.conf_params_xgboost, XgboostRegressionFinalParamConfig)

        # Ensure logger was called with appropriate messages
        assert mock_logger.call_count == 0


def test_check_load_confs_all_provided():
    conf_training = TrainingConfig()
    conf_xgboost = XgboostTuneParamsRegressionConfig()
    conf_params_xgboost = XgboostRegressionFinalParamConfig()

    model = XgboostModelRegression(
        class_problem="regression",
        conf_training=conf_training,
        conf_xgboost=conf_xgboost,
        conf_params_xgboost=conf_params_xgboost,
    )

    with patch("bluecast.ml_modelling.xgboost_regression.logging") as mock_logger:
        model.check_load_confs()

        # Check if all provided configs are used
        assert model.conf_training is conf_training
        assert model.conf_xgboost is conf_xgboost
        assert model.conf_params_xgboost is conf_params_xgboost

        # Ensure logger was called with appropriate messages
        assert mock_logger.call_count == 0


def test_xgboost_regression_fit_errors(synthetic_train_test_data_xgboost):
    model = XgboostModelRegression(
        class_problem="regression",
    )
    df_train, y_train, df_val, y_test = synthetic_train_test_data_xgboost
    with pytest.raises(ValueError):
        model.fit(df_train, df_val, y_train, y_test)



from typing import Union

import numpy as np
import pandas as pd


def absolute_error(
    y_true: Union[pd.Series, np.ndarray], y_hat: np.ndarray
) -> np.ndarray:
    """
    Calculate the absolute error per row.

    It calculates the absolute error between the predictions and the actual results.
    :param y_true: True values
    :param y_hat: Predicted values
    :return: Absolute error per row
    """
    if isinstance(y_true, pd.Series):
        return np.abs(y_hat - y_true.values)
    else:
        return np.abs(y_hat - y_true)


def normalized_error(
    y_true: Union[pd.Series, np.ndarray], y_hat: np.ndarray, scale: float
) -> np.ndarray:
    """
    Calculate the normalized error per row.

    The normalized error nonconformity measure is the absolute error divided
    by an estimate of the prediction error’s scale, such as the mean absolute error (MAE) or the
    standard deviation of the residuals. This measure can be used with any regression model and
    helps account for heteroscedasticity in the data.
    :param y_true: True values
    :param y_hat: Predicted values
    :param scale: Scale to normalize the error
    :return: Absolute error per row
    """
    if isinstance(y_true, pd.Series):
        return np.abs(y_hat - y_true.values) / scale
    else:
        return np.abs(y_hat - y_true) / scale



import numpy as np
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split

from bluecast.conformal_prediction.conformal_prediction import (
    ConformalPredictionWrapper,
)


def test_calibrate():
    # Generate some random data
    X, y = make_classification(
        n_samples=100, n_features=5, random_state=42, n_classes=2
    )
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a logistic regression model
    model = LogisticRegression(random_state=42)
    model.fit(X_train, y_train)

    # Create a ConformalPredictionWrapper instance
    wrapper = ConformalPredictionWrapper(model)

    # Calibrate the wrapper
    wrapper.calibrate(X_calibrate, y_calibrate)

    # Check that the nonconformity scores have been calculated
    assert isinstance(wrapper.nonconformity_scores, np.ndarray)
    assert len(wrapper.nonconformity_scores) == len(y_calibrate)

    # Check that the nonconformity scores are floats
    for score in wrapper.nonconformity_scores:
        assert isinstance(score, float)


def test_predict():
    # Generate some random data
    X, y = make_classification(
        n_samples=100, n_features=5, random_state=42, n_classes=2
    )
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a logistic regression model
    model = LogisticRegression(random_state=42)
    model.fit(X_train, y_train)

    # Create a ConformalPredictionWrapper instance
    wrapper = ConformalPredictionWrapper(model)

    # Make predictions
    y_pred = wrapper.predict(X_test)

    # Check that the predictions have the correct shape
    assert y_pred.shape == (len(X_test),)


def test_predict_proba():
    # Generate some random data
    X, y = make_classification(
        n_samples=100, n_features=5, random_state=42, n_classes=2
    )
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a logistic regression model
    model = LogisticRegression(random_state=42)
    model.fit(X_train, y_train)

    # Create a ConformalPredictionWrapper instance
    wrapper = ConformalPredictionWrapper(model)

    # Make predictions
    y_pred_proba = wrapper.predict_proba(X_test)

    # Check that the predictions have the correct shape
    assert y_pred_proba.shape == (len(X_test), 2)


def test_predict_interval():
    # Generate some random data
    X, y = make_classification(
        n_samples=100, n_features=5, random_state=42, n_classes=2
    )
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Train a logistic regression model
    model = LogisticRegression(random_state=42)
    model.fit(X_train, y_train)

    # Create a ConformalPredictionWrapper instance
    wrapper = ConformalPredictionWrapper(model)

    # Calibrate the wrapper
    wrapper.calibrate(X_calibrate, y_calibrate)

    # Make predictions
    y_pred_interval = wrapper.predict_interval(X_calibrate)

    # Check that the predictions have the correct shape
    assert y_pred_interval.shape == (len(X_calibrate), 2)


def test_predict_sets():
    # Generate some random data
    X, y = make_classification(
        n_samples=400, n_features=5, random_state=42, n_classes=2
    )
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X_train, y_train, test_size=0.2, random_state=42
    )

    # Train a logistic regression model
    model = LogisticRegression(random_state=42)
    model.fit(X_train, y_train)

    # Create a ConformalPredictionWrapper instance
    wrapper = ConformalPredictionWrapper(model)

    # Calibrate the wrapper
    wrapper.calibrate(X_calibrate, y_calibrate)

    # Make predictions
    alpha = 0.05
    y_pred_sets = wrapper.predict_sets(X_test, alpha=alpha)

    # Check that the predictions have the correct shape
    assert len(y_pred_sets) == len(X_test)

    # Check that each prediction set is a set
    for prediction_set in y_pred_sets:
        assert isinstance(prediction_set, set)

        # Check that each element in the prediction set is a tuple
        for element in prediction_set:
            # Check that each tuple has one or two elements
            assert element in [0, 1]

            # Check that each element in the tuple is an integer
            assert isinstance(element, int)

    # Count correct predictions
    correct_predictions = sum(
        1 for pred_set, true_value in zip(y_pred_sets, y_test) if true_value in pred_set
    )

    # Calculate percentage
    assert correct_predictions / len(y_test) >= 1 - alpha

    # Make predictions
    alpha = 0.01
    y_pred_sets = wrapper.predict_sets(X_test, alpha=alpha)
    correct_predictions = sum(
        1 for pred_set, true_value in zip(y_pred_sets, y_test) if true_value in pred_set
    )

    assert correct_predictions / len(y_test) >= 1 - alpha



from unittest.mock import patch

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostTuneParamsConfig,
)
from bluecast.ml_modelling.parameter_tuning_utils import get_params_based_on_device


def test_get_params_based_on_device_auto_gpu():
    conf_training = TrainingConfig(autotune_on_device="auto")
    conf_params_xgboost = XgboostFinalParamConfig()
    conf_xgboost = XgboostTuneParamsConfig(tree_method=["exact"])

    mock_gpu_support = {"device": "cuda"}

    with patch(
        "bluecast.general_utils.general_utils.check_gpu_support",
        return_value=mock_gpu_support,
    ):
        result = get_params_based_on_device(
            conf_training, conf_params_xgboost, conf_xgboost
        )

    assert result["device"] == "cpu"


def test_get_params_based_on_device_cpu():
    conf_training = TrainingConfig(autotune_on_device="cpu")
    conf_params_xgboost = XgboostFinalParamConfig()
    conf_xgboost = XgboostTuneParamsConfig(tree_method=["exact"])

    with patch("bluecast.general_utils.general_utils.check_gpu_support"):
        result = get_params_based_on_device(
            conf_training, conf_params_xgboost, conf_xgboost
        )

    assert result["device"] == "cpu"
    assert result["tree_method"] == "exact"
    assert "exact" in conf_xgboost.tree_method


def test_get_params_based_on_device_gpu_with_exact_tree_method():
    conf_training = TrainingConfig(autotune_on_device="gpu")
    conf_params_xgboost = XgboostFinalParamConfig()
    conf_xgboost = XgboostTuneParamsConfig(tree_method=["exact", "hist"])

    mock_gpu_support = {"device": "cuda"}

    with patch(
        "bluecast.general_utils.general_utils.check_gpu_support",
        return_value=mock_gpu_support,
    ):
        result = get_params_based_on_device(
            conf_training, conf_params_xgboost, conf_xgboost
        )

    assert result["device"] == "cpu"


def test_get_params_based_on_device_gpu_without_exact_tree_method():
    conf_training = TrainingConfig(autotune_on_device="gpu")
    conf_params_xgboost = XgboostFinalParamConfig()
    conf_xgboost = XgboostTuneParamsConfig(tree_method=["hist"])

    mock_gpu_support = {"device": "cuda"}

    with patch(
        "bluecast.general_utils.general_utils.check_gpu_support",
        return_value=mock_gpu_support,
    ):
        result = get_params_based_on_device(
            conf_training, conf_params_xgboost, conf_xgboost
        )

    assert result["device"] == "cpu"



import logging
from typing import Any, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import RepeatedKFold

from bluecast.blueprints.cast_regression import BlueCastRegression
from bluecast.config.training_config import (
    TrainingConfig,
    XgboostRegressionFinalParamConfig,
    XgboostTuneParamsRegressionConfig,
)
from bluecast.conformal_prediction.conformal_prediction_regression import (
    ConformalPredictionRegressionWrapper,
)
from bluecast.evaluation.eval_metrics import RegressionEvalWrapper
from bluecast.experimentation.tracking import ExperimentTracker
from bluecast.ml_modelling.xgboost import XgboostModel
from bluecast.preprocessing.custom import CustomPreprocessing
from bluecast.preprocessing.feature_selection import BoostaRootaWrapper


class BlueCastCVRegression:
    """Wrapper to train and predict multiple blueCast instances.

    Check the BlueCast class documentation for additional parameter details.
    A custom splitter can be provided.

    :param :class_problem: Takes a string containing the class problem type. At the moment "regression" only.
    :param :target_column: Takes a string containing the name of the target column.
    :param :cat_columns: Takes a list of strings containing the names of the categorical columns. If not provided,
        BlueCast will infer these automatically.
    :param :date_columns: Takes a list of strings containing the names of the date columns. If not provided,
        BlueCast will infer these automatically.
    :param :time_split_column: Takes a string containing the name of the time split column. If not provided,
        BlueCast will not split the data by time or order, but do a random split instead.
    :param :ml_model: Takes an instance of a XgboostModelRegression class. If not provided, BlueCast will instantiate one.
        This is an API to pass any model class. Inherit the baseclass from ml_modelling.base_model.BaseModel.
    :param custom_in_fold_preprocessor: Takes an instance of a CustomPreprocessing class. Allows users to eeecute
        preprocessing after the train test split within cv folds. This will be executed only if precise_cv_tuning in
        the conf_Training is True. Custom ML models need to implement this themselves. This step is only useful when
        the proprocessing step has a high chance of overfitting otherwise (i.e: oversampling techniques).
    :param custom_preprocessor: Takes an instance of a CustomPreprocessing class. Allows users to inject custom
        preprocessing steps which take place right after the train test spit.
    :param custom_last_mile_computation: Takes an instance of a CustomPreprocessing class. Allows users to inject custom
        preprocessing steps which take place right before the model training.
    :param experiment_tracker: Takes an instance of an ExperimentTracker class. If not provided this will be initialized
        automatically.
    :param single_fold_eval_metric_func: Takes a function which calculates the evaluation metric for a single fold.
       Default is mean_squared_error. This function is used to calculate the evaluation metric for each fold during
       hyperparameter tuning when hyperparameter_tuning_rounds = 1 (default). Lower must be better.
    """

    def __init__(
        self,
        class_problem: Literal["regression"] = "regression",
        cat_columns: Optional[List[Union[str, float, int]]] = None,
        stratifier: Optional[Any] = None,
        conf_training: Optional[TrainingConfig] = None,
        conf_xgboost: Optional[XgboostTuneParamsRegressionConfig] = None,
        conf_params_xgboost: Optional[XgboostRegressionFinalParamConfig] = None,
        experiment_tracker: Optional[ExperimentTracker] = None,
        custom_in_fold_preprocessor: Optional[CustomPreprocessing] = None,
        custom_last_mile_computation: Optional[CustomPreprocessing] = None,
        custom_preprocessor: Optional[CustomPreprocessing] = None,
        custom_feature_selector: Optional[
            Union[BoostaRootaWrapper, CustomPreprocessing]
        ] = None,
        ml_model: Optional[Union[XgboostModel, Any]] = None,
        single_fold_eval_metric_func: Optional[RegressionEvalWrapper] = None,
    ):
        self.class_problem = class_problem
        self.conf_xgboost = conf_xgboost
        self.conf_training = conf_training
        self.conf_params_xgboost = conf_params_xgboost
        self.custom_in_fold_preprocessor = custom_in_fold_preprocessor
        self.custom_preprocessor = custom_preprocessor
        self.custom_feature_selector = custom_feature_selector
        self.custom_last_mile_computation = custom_last_mile_computation
        self.bluecast_models: List[BlueCastRegression] = []
        self.stratifier = stratifier
        self.ml_model = ml_model
        self.single_fold_eval_metric_func = single_fold_eval_metric_func
        self.conformal_prediction_wrapper: Optional[
            ConformalPredictionRegressionWrapper
        ] = None

        if not cat_columns:
            self.cat_columns = []
        else:
            self.cat_columns = cat_columns

        if experiment_tracker:
            self.experiment_tracker = experiment_tracker
        else:
            self.experiment_tracker = ExperimentTracker()

        if not self.conf_params_xgboost:
            self.conf_params_xgboost = XgboostRegressionFinalParamConfig()

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.conf_xgboost:
            self.conf_xgboost = XgboostTuneParamsRegressionConfig()

        if not self.single_fold_eval_metric_func:
            self.single_fold_eval_metric_func = RegressionEvalWrapper(
                higher_is_better=False,
                metric_func=mean_squared_error,
                metric_name="Mean squared error",
                **{"squared": False},
            )

    def prepare_data(
        self, df: pd.DataFrame, target: str
    ) -> Tuple[pd.DataFrame, pd.Series]:
        df = df.reset_index(drop=True)
        y = df[target]
        X = df.drop(target, axis=1)
        return X, y

    def show_oof_scores(self, metric: str = "RMSE") -> Tuple[float, float]:
        """
        Show out of fold scores.

        When calling BlueCastCVRegression's fit_eval function multiple BlueCastRegression
        instances are called and each of them predicts on unseen/oof data.

        This function collects these scores and return mean and average of them.

        :param metric: String indicating which metric shall be returned.
        :return: Tuple with (mean, std) of oof scores
        """
        all_metrics = []
        for bluecast_instance in self.bluecast_models:
            if bluecast_instance.eval_metrics:
                score = bluecast_instance.eval_metrics.get(metric)
                all_metrics.append(score)

        score_mean = np.asarray(all_metrics).mean()
        score_std = np.asarray(all_metrics).std()
        logging.info(
            f"The mean out of fold {metric} score is {score_mean} with an std of {score_std}"
        )
        return score_mean, score_std

    def fit(self, df: pd.DataFrame, target_col: str) -> None:
        """Fit multiple BlueCastRegression instances on different data splits.

        Input df is expected the target column."""
        X, y = self.prepare_data(df, target_col)

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.stratifier:
            self.stratifier = RepeatedKFold(
                n_splits=self.conf_training.bluecast_cv_train_n_model[0],
                n_repeats=self.conf_training.bluecast_cv_train_n_model[1],
                random_state=self.conf_training.global_random_state,
            )

        for fn, (trn_idx, val_idx) in enumerate(self.stratifier.split(X, y)):
            X_train, X_val = X.iloc[trn_idx], X.iloc[val_idx]
            y_train, y_val = y.iloc[trn_idx], y.iloc[val_idx]
            x_train = pd.concat([X_train, X_val], ignore_index=True)
            y_train = pd.concat([y_train, y_val], ignore_index=True)

            X_train = x_train.reset_index(drop=True)
            y_train = y_train.reset_index(drop=True)
            X_train[target_col] = y_train.values

            self.conf_training.global_random_state += (
                self.conf_training.increase_random_state_in_bluecast_cv_by
            )
            logging.info(
                f"Start fitting model number {fn} with random seed {self.conf_training.global_random_state}"
            )

            automl = BlueCastRegression(
                class_problem=self.class_problem,
                cat_columns=self.cat_columns,
                conf_training=self.conf_training,
                conf_xgboost=self.conf_xgboost,
                conf_params_xgboost=self.conf_params_xgboost,
                experiment_tracker=self.experiment_tracker,
                custom_in_fold_preprocessor=self.custom_in_fold_preprocessor,
                custom_preprocessor=self.custom_preprocessor,
                custom_feature_selector=self.custom_feature_selector,
                custom_last_mile_computation=self.custom_last_mile_computation,
                ml_model=self.ml_model,
                single_fold_eval_metric_func=self.single_fold_eval_metric_func,
            )
            automl.fit(X_train, target_col=target_col)
            self.bluecast_models.append(automl)

            # overwrite experiment tracker to pass it into next iteration
            self.experiment_tracker = automl.experiment_tracker

    def fit_eval(self, df: pd.DataFrame, target_col: str) -> Tuple[float, float]:
        """Fit multiple BlueCastRegression instances on different data splits.

        Input df is expected the target column. Evaluation is executed on out-of-fold dataset
        in each split.
        :param df: Pandas DataFrame that includes the target column
        :param target_col: String indicating the name of the target column
        :returns Tuple of (oof_mean, oof_std) with scores on unseen data during eval
        """
        X, y = self.prepare_data(df, target_col)

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.stratifier:
            self.stratifier = RepeatedKFold(
                n_splits=self.conf_training.bluecast_cv_train_n_model[0],
                n_repeats=self.conf_training.bluecast_cv_train_n_model[1],
                random_state=self.conf_training.global_random_state,
            )

        for fn, (trn_idx, val_idx) in enumerate(self.stratifier.split(X, y)):
            X_train, X_val = X.iloc[trn_idx], X.iloc[val_idx]
            y_train, y_val = y.iloc[trn_idx], y.iloc[val_idx]

            X_train.loc[:, target_col] = y_train

            self.conf_training.global_random_state += (
                self.conf_training.increase_random_state_in_bluecast_cv_by
            )
            logging.info(
                f"Start fitting model number {fn} with random seed {self.conf_training.global_random_state}"
            )

            automl = BlueCastRegression(
                class_problem=self.class_problem,
                cat_columns=self.cat_columns,
                conf_training=self.conf_training,
                conf_xgboost=self.conf_xgboost,
                conf_params_xgboost=self.conf_params_xgboost,
                experiment_tracker=self.experiment_tracker,
                custom_in_fold_preprocessor=self.custom_in_fold_preprocessor,
                custom_preprocessor=self.custom_preprocessor,
                custom_feature_selector=self.custom_feature_selector,
                custom_last_mile_computation=self.custom_last_mile_computation,
                ml_model=self.ml_model,
                single_fold_eval_metric_func=self.single_fold_eval_metric_func,
            )
            automl.fit_eval(X_train, X_val, y_val, target_col=target_col)
            self.bluecast_models.append(automl)

            # overwrite experiment tracker to pass it into next iteration
            self.experiment_tracker = automl.experiment_tracker

        oof_mean, oof_std = self.show_oof_scores()
        return oof_mean, oof_std

    def predict(
        self,
        df: pd.DataFrame,
        return_sub_models_preds: bool = False,
        save_shap_values: bool = False,
        mean_type: Literal[
            "arithmetic", "median", "geometric", "harmonic"
        ] = "arithmetic",
    ) -> Union[pd.DataFrame, pd.Series]:
        """Predict on unseen data using multiple trained BlueCastRegression instances.

        :param df: Pandas DataFrame with unseen data
        :param return_sub_models_preds: If true will return a DataFrame with the predictions of each model
            stored in separate columns.
        :param save_shap_values: If True, calculates and saves shap values, so they can be used to plot
            waterfall plots for selected rows o demand.
        :param mean_type: String indicating the type of mean to be used to blend the predictions of the sub models.
            Possible values are 'arithmetic', 'geometric' and 'harmonic' (default='arithmetic').
        """
        or_cols = df.columns
        pred_cols: list[str] = []
        result_df = pd.DataFrame()  # Create an empty DataFrame to store results

        for fn, pipeline in enumerate(self.bluecast_models):
            y_preds = pipeline.predict(
                df.loc[:, or_cols], save_shap_values=save_shap_values
            )
            result_df[f"preds_{fn}"] = y_preds
            pred_cols.append(f"preds_{fn}")

        if return_sub_models_preds:
            return result_df
        else:
            if mean_type == "arithmetic":
                return result_df.mean(axis=1)
            elif mean_type == "geometric":
                return np.exp(np.log(result_df.prod(axis=1)) / result_df.notna().sum(1))
            elif mean_type == "harmonic":
                return len(pred_cols) / np.sum(1 / result_df, axis=1)
            elif mean_type == "median":
                return result_df.median(axis=1)
            else:
                return result_df.mean(axis=1)

    def calibrate(
        self, x_calibration: pd.DataFrame, y_calibration: pd.Series, **kwargs
    ) -> None:
        """Calibrate the model.

        Via this function the nonconformity measures are taken and used to predict prediction intervals vis the
        predict_interval function. Used is the mean prediction of all sub models.
        :param: x_calibration: Pandas DataFrame without target column, that has not been seen by the model during
            training.
        :param y_calibration: Pandas Series holding the target value, hat has not been seen by the model during
            training.
        """
        self.conformal_prediction_wrapper = ConformalPredictionRegressionWrapper(
            self, **kwargs
        )
        self.conformal_prediction_wrapper.calibrate(x_calibration, y_calibration)

    def predict_interval(self, df: pd.DataFrame, alphas: List[float]) -> pd.DataFrame:
        """Create prediction intervals based on a certain confidence levels.

        Conformal prediction guarantees, that the correct value is present in the prediction band with a probability of
        1 - alpha.
        :param df: Pandas DataFrame holding unseen data
        :param alphas: List of floats indicating the desired confidence levels.
        :returns A Pandas DataFrame with  sorted columns 'alpha_XX_low' (alpha) and 'alpha_XX_high' (1 - alpha) for each
            alpha in the provided list of alphas. To obtain the mean prediction call the 'predict' method.
        """
        if self.conformal_prediction_wrapper:
            pred_interval = self.conformal_prediction_wrapper.predict_interval(
                df, alphas=alphas
            )
            return pred_interval
        else:
            raise ValueError(
                """This instance has not been calibrated yet. Make use of calibrate to fit the
            ConformalPredictionWrapper."""
            )



from collections import namedtuple

import pandas as pd
import pytest

from bluecast.preprocessing.feature_creation import (
    AddRowLevelAggFeatures,
    FeatureClusteringScorer,
)


@pytest.fixture
def sample_dataframe():
    data = {
        "A": [1, 2, 3, 4],
        "B": [5, 6, 7, 8],
        "C": [9, 10, 11, 12],
        "target": [0, 1, 0, 1],
    }
    return pd.DataFrame(data)


def test_initialization():
    aggregator = AddRowLevelAggFeatures()
    assert aggregator.original_features == []


def test_get_original_features(sample_dataframe):
    aggregator = AddRowLevelAggFeatures()
    aggregator.get_original_features(sample_dataframe, "target")
    assert aggregator.original_features == ["A", "B", "C"]


def test_add_row_level_mean(sample_dataframe):
    aggregator = AddRowLevelAggFeatures()
    df = aggregator.add_row_level_mean(sample_dataframe, ["A", "B", "C"], "row_mean")
    expected_mean = sample_dataframe[["A", "B", "C"]].mean(axis=1)
    pd.testing.assert_series_equal(
        df["row_mean"], expected_mean, check_names=False, check_like=True
    )


def test_add_row_level_std(sample_dataframe):
    aggregator = AddRowLevelAggFeatures()
    df = aggregator.add_row_level_std(sample_dataframe, ["A", "B", "C"], "row_std")
    expected_std = sample_dataframe[["A", "B", "C"]].std(axis=1)
    pd.testing.assert_series_equal(df["row_std"], expected_std, check_names=False)


def test_add_row_level_min(sample_dataframe):
    aggregator = AddRowLevelAggFeatures()
    df = aggregator.add_row_level_min(sample_dataframe, ["A", "B", "C"], "row_min")
    expected_min = sample_dataframe[["A", "B", "C"]].min(axis=1)
    pd.testing.assert_series_equal(df["row_min"], expected_min, check_names=False)


def test_add_row_level_max(sample_dataframe):
    aggregator = AddRowLevelAggFeatures()
    df = aggregator.add_row_level_max(sample_dataframe, ["A", "B", "C"], "row_max")
    expected_max = sample_dataframe[["A", "B", "C"]].max(axis=1)
    pd.testing.assert_series_equal(df["row_max"], expected_max, check_names=False)


def test_add_row_level_sum(sample_dataframe):
    aggregator = AddRowLevelAggFeatures()
    df = aggregator.add_row_level_sum(sample_dataframe, ["A", "B", "C"], "row_sum")
    expected_sum = sample_dataframe[["A", "B", "C"]].sum(axis=1)
    pd.testing.assert_series_equal(df["row_sum"], expected_sum, check_names=False)


def test_add_row_level_agg_features(sample_dataframe):
    aggregator = AddRowLevelAggFeatures()
    df = aggregator.add_row_level_agg_features(sample_dataframe, "target")

    # Verify if the correct columns are added
    expected_mean = sample_dataframe[["A", "B", "C"]].mean(axis=1)
    expected_std = sample_dataframe[["A", "B", "C"]].std(axis=1)
    expected_min = sample_dataframe[["A", "B", "C"]].min(axis=1)
    expected_max = sample_dataframe[["A", "B", "C"]].max(axis=1)
    expected_sum = sample_dataframe[["A", "B", "C"]].sum(axis=1)

    pd.testing.assert_series_equal(df["row_mean"], expected_mean, check_names=False)
    pd.testing.assert_series_equal(df["row_std"], expected_std, check_names=False)
    pd.testing.assert_series_equal(df["row_min"], expected_min, check_names=False)
    pd.testing.assert_series_equal(df["row_max"], expected_max, check_names=False)
    pd.testing.assert_series_equal(df["row_sum"], expected_sum, check_names=False)


@pytest.fixture
def synthetic_data():
    test_df = pd.DataFrame(
        {
            "customer_id": [f"M{i}" for i in range(100)],
            "recency": [i for i in range(100)][::-1],  # inverse order
            "frequency": [i for i in range(100)],
            "monetary": [i for i in range(100)],
        }
    )
    return test_df


def test_higher_is_better(synthetic_data):
    """Test if higher is better than lower param works"""
    test_df = synthetic_data

    cluster_setting = namedtuple(
        "cluster_settings", ("nb_clusters", "higher_is_better")
    )

    cluster_settings = {
        "recency": cluster_setting(2, False),
        "frequency": cluster_setting(3, True),
        "monetary": cluster_setting(2, True),
    }

    rfm_c = FeatureClusteringScorer(cluster_settings)
    cluster_results = rfm_c.fit_predict_cluster(test_df, keep_original_features=False)
    assert cluster_results.head(1)["recency"].values[0] == 1
    assert cluster_results.tail(1)["recency"].values[0] == 2

    assert cluster_results.head(1)["frequency"].values[0] == 1
    assert cluster_results.tail(1)["frequency"].values[0] == 3

    assert cluster_results.head(1)["monetary"].values[0] == 1
    assert cluster_results.tail(1)["monetary"].values[0] == 2

    cluster_settings = {
        "recency": cluster_setting(2, True),
        "frequency": cluster_setting(3, False),
        "monetary": cluster_setting(2, False),
    }

    rfm_c = FeatureClusteringScorer(cluster_settings)
    cluster_results = rfm_c.fit_predict_cluster(test_df, keep_original_features=False)
    assert cluster_results.head(1)["recency"].values[0] == 2
    assert cluster_results.tail(1)["recency"].values[0] == 1

    assert cluster_results.head(1)["frequency"].values[0] == 3
    assert cluster_results.tail(1)["frequency"].values[0] == 1

    assert cluster_results.head(1)["monetary"].values[0] == 2
    assert cluster_results.tail(1)["monetary"].values[0] == 1


def test_keep_original_features(synthetic_data):
    """Test if keep original features works"""
    test_df = synthetic_data

    cluster_setting = namedtuple(
        "cluster_settings", ("nb_clusters", "higher_is_better")
    )

    cluster_settings = {
        "recency": cluster_setting(2, False),
        "frequency": cluster_setting(3, True),
        "monetary": cluster_setting(2, True),
    }

    rfm_c = FeatureClusteringScorer(cluster_settings)
    cluster_results = rfm_c.fit_predict_cluster(test_df, keep_original_features=False)
    assert (
        len(cluster_results.columns) == len(cluster_settings) + 1
    )  # +1 for total score

    cluster_results = rfm_c.predict_cluster(test_df, keep_original_features=False)
    assert (
        len(cluster_results.columns) == len(cluster_settings) + 1
    )  # +1 for total score

    rfm_c = FeatureClusteringScorer(cluster_settings)
    cluster_results = rfm_c.fit_predict_cluster(test_df, keep_original_features=True)
    assert (
        len(cluster_results.columns) == len(cluster_settings) + len(test_df.columns) + 1
    )  # +1 for total score

    assert "recency" in cluster_results.columns
    assert "frequency" in cluster_results.columns
    assert "monetary" in cluster_results.columns
    assert "total_score" in cluster_results.columns
    assert "customer_id_original" in cluster_results.columns
    assert "recency_original" in cluster_results.columns
    assert "frequency_original" in cluster_results.columns
    assert "monetary_original" in cluster_results.columns

    cluster_results = rfm_c.predict_cluster(test_df, keep_original_features=True)
    assert (
        len(cluster_results.columns) == len(cluster_settings) + len(test_df.columns) + 1
    )  # +1 for total score

    assert "recency" in cluster_results.columns
    assert "frequency" in cluster_results.columns
    assert "monetary" in cluster_results.columns
    assert "total_score" in cluster_results.columns
    assert "customer_id_original" in cluster_results.columns
    assert "recency_original" in cluster_results.columns
    assert "frequency_original" in cluster_results.columns
    assert "monetary_original" in cluster_results.columns


def test_changing_features(synthetic_data):
    """Test if can use custom features"""
    test_df = synthetic_data
    test_df["loyalty"] = [i for i in range(100)]

    cluster_setting = namedtuple(
        "cluster_settings", ("nb_clusters", "higher_is_better")
    )

    cluster_settings = {
        "recency": cluster_setting(2, False),
        "frequency": cluster_setting(3, True),
        "monetary": cluster_setting(2, True),
        "loyalty": cluster_setting(4, True),
    }

    rfm_c = FeatureClusteringScorer(cluster_settings)
    cluster_results = rfm_c.fit_predict_cluster(test_df, keep_original_features=True)

    assert "loyalty" in cluster_results.columns
    assert "loyalty_original" in cluster_results.columns
    assert cluster_results.head(1)["loyalty"].values[0] == 1
    assert cluster_results.tail(1)["loyalty"].values[0] == 4
    assert cluster_results["total_score"].max() == 11

    cluster_results = rfm_c.predict_cluster(test_df, keep_original_features=True)

    assert "loyalty" in cluster_results.columns
    assert "loyalty_original" in cluster_results.columns
    assert cluster_results.head(1)["loyalty"].values[0] == 1
    assert cluster_results.tail(1)["loyalty"].values[0] == 4
    assert cluster_results["total_score"].max() == 11



from unittest.mock import MagicMock, Mock

import optuna

from bluecast.config.training_config import XgboostTuneParamsConfig
from bluecast.ml_modelling.parameter_tuning_utils import (
    update_hyperparam_space_after_nth_trial,
    update_params_based_on_tree_method,
)

# Mocking XgboostTuneParamsConfig for testing
mock_xgboost_params = XgboostTuneParamsConfig(
    tree_method=["hist", "gpu_hist"],
    booster=["gbtree", "gblinear"],
    grow_policy=["depthwise", "lossguide"],
)

# Mocking optuna.Trial for testing
mock_trial = Mock(spec=optuna.Trial)


def setup_mock_trial(trial: Mock, booster: str, tree_method: str, grow_policy: str):
    trial.suggest_categorical.side_effect = [tree_method, booster, grow_policy]


def test_update_params_based_on_tree_method_gbtree():
    param = {
        "booster": "gbtree",
        "tree_method": "auto",  # Initial value for testing
        "max_depth": 5,
        "min_child_weight": 1,
        "subsample": 0.8,
        "colsample_bytree": 0.8,
        "colsample_bylevel": 0.8,
        "gamma": 0.2,
    }

    setup_mock_trial(mock_trial, "gbtree", "hist", "depthwise")

    updated_param = update_params_based_on_tree_method(
        param, mock_trial, mock_xgboost_params
    )

    assert updated_param["booster"] == "gbtree"
    assert updated_param["tree_method"] == "hist"
    assert updated_param["grow_policy"] == "depthwise"
    assert "max_depth" in updated_param
    assert "min_child_weight" in updated_param
    assert "subsample" in updated_param
    assert "colsample_bytree" in updated_param
    assert "colsample_bylevel" in updated_param
    assert "gamma" in updated_param


def test_update_params_based_on_tree_method_gblinear():
    param = {
        "booster": "gblinear",
        "max_depth": 5,
        "min_child_weight": 1,
        "subsample": 0.8,
        "colsample_bytree": 0.8,
        "colsample_bylevel": 0.8,
        "gamma": 0.2,
        "tree_method": "auto",
    }

    setup_mock_trial(mock_trial, "gblinear", "hist", "depthwise")

    updated_param = update_params_based_on_tree_method(
        param, mock_trial, mock_xgboost_params
    )

    assert updated_param["booster"] == "gblinear"
    assert "tree_method" not in updated_param
    assert "max_depth" not in updated_param
    assert "min_child_weight" not in updated_param
    assert "subsample" not in updated_param
    assert "colsample_bytree" not in updated_param
    assert "colsample_bylevel" not in updated_param
    assert "gamma" not in updated_param


def test_update_hyperparam_space_after_nth_trial():
    # Create a config object
    conf_xgboost = XgboostTuneParamsConfig()

    trial_data = [
        (1, 0.001, 0.3, 0.5, 100.0),
        (25, 0.05, 0.25, 1.0, 10.0),
        (50, 0.001, 0.3, 0.5, 100.0),
        (75, 0.05, 0.25, 1.0, 10.0),
        (100, 0.001, 0.3, 0.5, 100.0),
    ]

    for (
        trial_number,
        expected_eta_min,
        expected_eta_max,
        expected_sub_sample_min,
        expected_min_child_weight_max,
    ) in trial_data:
        trial = MagicMock()
        trial.number = trial_number
        # Call the function
        conf_xgboost = update_hyperparam_space_after_nth_trial(
            trial, conf_xgboost, nth_trial=25
        )
        print(
            trial_number,
            trial_number % 25 * 2,
            trial_number % 25,
            conf_xgboost.eta_min,
            expected_eta_min,
        )

        # Assertions
        assert conf_xgboost.eta_min == expected_eta_min
        assert conf_xgboost.eta_max == expected_eta_max
        assert conf_xgboost.sub_sample_min == expected_sub_sample_min
        assert conf_xgboost.min_child_weight_max == expected_min_child_weight_max



"""
Onehot encoding is a method to encode categorical features. It is an unsupervised encoding technique.
It is not recommended for features with high cardinality.

The onehot encoding technique is implemented in the category_encoders library. The library offers a variety of
different encoding techniques. The onehot encoding technique is implemented in the OneHotEncoder class.
"""

import logging
from typing import Dict, List, Union

import pandas as pd
from category_encoders import OneHotEncoder


class OneHotCategoryEncoder:
    """Onehot encode categorical features."""

    def __init__(
        self,
        cat_columns: List[Union[str, float, int]],
        target_col: Union[str, float, int],
    ):
        self.encoders: Dict[str, OneHotEncoder] = {}
        self.prediction_mode: bool = False
        self.cat_columns = cat_columns
        self.target_col = target_col

    def fit_transform(self, x: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        """Fit onehot encoder and transform column."""
        logging.info("Start fitting binary target encoder.")
        if self.target_col in self.cat_columns:
            self.cat_columns.remove(self.target_col)

        enc = OneHotEncoder(
            use_cat_names=True,  # drop_invariant=True, use_cat_names=True
        )
        encoded_cats = enc.fit_transform(x[self.cat_columns], y)
        x_new = x.drop(
            self.cat_columns, axis=1
        ).copy()  # copy against high fragmentation
        x_new = self.append_encoded_columns(x_new, encoded_cats)

        self.encoders["onehot_encoder_all_cols"] = enc
        return x_new.copy()  # copy against high fragmentation

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        """Transform categories based on already trained encoder."""
        logging.info("Start transforming categories with binary target encoder.")
        enc = self.encoders["onehot_encoder_all_cols"]
        encoded_cats = enc.transform(x[self.cat_columns])
        x = x.drop(self.cat_columns, axis=1)
        x = self.append_encoded_columns(x, encoded_cats)
        return x

    def append_encoded_columns(
        self, x: pd.DataFrame, encoded_cats: pd.DataFrame
    ) -> pd.DataFrame:
        """Append encoded columns to the DataFrame."""
        x[encoded_cats.columns.to_list()] = encoded_cats
        x[encoded_cats.columns.to_list()] = x[encoded_cats.columns.to_list()].astype(
            int
        )
        # Set all new columns to -1 where all values are 0
        mask_all_zero = (x[encoded_cats.columns.to_list()] == 0).all(axis=1)
        x.loc[mask_all_zero, encoded_cats.columns.to_list()] = -1
        return x



from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.preprocessing import MinMaxScaler


class AddRowLevelAggFeatures:
    def __init__(self):
        self.original_features: List[Union[str, int, float]] = []

    def get_original_features(
        self, df: pd.DataFrame, target_col: Optional[str]
    ) -> None:
        if isinstance(target_col, str):
            self.original_features = df.drop(target_col, axis=1).columns.to_list()
        else:
            self.original_features = df.columns.to_list()

    def add_row_level_mean(
        self,
        df: pd.DataFrame,
        feature_to_agg: List[Union[str, int, float]],
        agg_col_name: str = "row_mean",
    ) -> pd.DataFrame:
        """
        Add row level mean of features to a dataframe.

        :param df: Pandas DataFrame holding all features.
        :param feature_to_agg: List of column names indicating which features to aggregate.
        :param agg_col_name: Name of the new column.
        :return: Original Pandas DataFrame with added row level means.
        """
        df[agg_col_name] = df[feature_to_agg].mean(axis=1)
        return df

    def add_row_level_std(
        self,
        df: pd.DataFrame,
        feature_to_agg: List[Union[str, int, float]],
        agg_col_name: str = "row_std",
    ) -> pd.DataFrame:
        """
        Add row level standard deviation of features to a dataframe.

        :param df: Pandas DataFrame holding all features.
        :param feature_to_agg: List of column names indicating which features to aggregate.
        :param agg_col_name: Name of the new column.
        :return: Original Pandas DataFrame with added row level means.
        """
        df[agg_col_name] = df[feature_to_agg].std(axis=1)
        return df

    def add_row_level_min(
        self,
        df: pd.DataFrame,
        feature_to_agg: List[Union[str, int, float]],
        agg_col_name: str = "row_min",
    ) -> pd.DataFrame:
        """
        Add row level min of features to a dataframe.

        :param df: Pandas DataFrame holding all features.
        :param feature_to_agg: List of column names indicating which features to aggregate.
        :param agg_col_name: Name of the new column.
        :return: Original Pandas DataFrame with added row level means.
        """
        df[agg_col_name] = df[feature_to_agg].min(axis=1)
        return df

    def add_row_level_max(
        self,
        df: pd.DataFrame,
        feature_to_agg: List[Union[str, int, float]],
        agg_col_name: str = "row_max",
    ) -> pd.DataFrame:
        """
        Add row level max of features to a dataframe.

        :param df: Pandas DataFrame holding all features.
        :param feature_to_agg: List of column names indicating which features to aggregate.
        :param agg_col_name: Name of the new column.
        :return: Original Pandas DataFrame with added row level means.
        """
        df[agg_col_name] = df[feature_to_agg].max(axis=1)
        return df

    def add_row_level_sum(
        self,
        df: pd.DataFrame,
        feature_to_agg: List[Union[str, int, float]],
        agg_col_name: str = "row_sum",
    ) -> pd.DataFrame:
        """
        Add row level sum of features to a dataframe.

        :param df: Pandas DataFrame holding all features.
        :param feature_to_agg: List of column names indicating which features to aggregate.
        :param agg_col_name: Name of the new column.
        :return: Original Pandas DataFrame with added row level means.
        """
        df[agg_col_name] = df[feature_to_agg].sum(axis=1)
        return df

    def add_row_level_agg_features(
        self, df: pd.DataFrame, target_col: Optional[str] = None
    ) -> pd.DataFrame:
        self.get_original_features(df, target_col)
        df = self.add_row_level_mean(df, self.original_features)
        df = self.add_row_level_std(df, self.original_features)
        df = self.add_row_level_min(df, self.original_features)
        df = self.add_row_level_max(df, self.original_features)
        df = self.add_row_level_sum(df, self.original_features)
        return df


class FeatureClusteringScorer:
    def __init__(
        self,
        cluster_settings: Dict[str, Any],
        random_state: int = 25,
    ):
        self.random_state = random_state  # control randomness
        self.cluster_settings = cluster_settings  # settings for each feature
        self.scalers: Dict[str, MinMaxScaler] = {}  # storing scalers per feature
        self.cluster_classes: Dict[str, KMeans] = {}  # storing Kmeans class per feature
        self.cluster_mappings: Dict[str, Dict[int, int]] = (
            {}
        )  # storing reindex mapping for cluster ids

    def _fit_reindex_clusters_by_mean(
        self, temp_df: pd.DataFrame, feature_name: str, higher_is_better: bool = True
    ) -> np.ndarray:
        """
        Fix cluster indices.

        Cluster indices do not follow the order of the original feature (i.e. highest value might nbe cluster 0).
        This function reindexes the cluster idx, so the total value make sense.

        :param temp_df: DataFrame containing two columns: the 'cluster' and the original feature
        :param feature_name: String indicating the name of the original feature.
        :param higher_is_better: Boolean indicating if the cluster index should raise with increasing values of the
            original feature.
        :return: Nmpy array with corrected cluster indices
        """
        # Calculate the mean of the feature values for each cluster
        cluster_means = temp_df.groupby("cluster")[feature_name].mean()

        # Sort the clusters by their mean values
        if higher_is_better:
            sorted_clusters = cluster_means.sort_values(ascending=True).index
        else:
            sorted_clusters = cluster_means.sort_values(ascending=False).index

        # Create a mapping from old cluster IDs to new cluster IDs
        self.cluster_mappings[feature_name] = {
            old: new for new, old in enumerate(sorted_clusters)
        }

        # Apply the mapping to reindex the cluster IDs in the DataFrame
        temp_df["cluster"] = temp_df["cluster"].map(self.cluster_mappings[feature_name])

        return temp_df["cluster"].values

    def _predict_reindex_clusters_by_mean(
        self, temp_df: pd.DataFrame, feature_name: str
    ) -> np.ndarray:
        """
        Fix cluster indices.

        Cluster indices do not follow the order of the original feature (i.e. highest value might nbe cluster 0).
        This function reindexes the cluster idx, so the total value make sense.

        :param temp_df: DataFrame containing two columns: the 'cluster' and the original feature
        :param feature_name: String indicating the name of the original feature.
        :return: Nmpy array with corrected cluster indices
        """
        # Apply the mapping to reindex the cluster IDs in the DataFrame
        temp_df["cluster"] = temp_df["cluster"].map(self.cluster_mappings[feature_name])

        return temp_df["cluster"].values

    def _fit_cluster_feature(
        self,
        df: pd.DataFrame,
        feature_name: str,
        nb_clusters: int,
        higher_is_better: bool,
    ) -> np.ndarray:
        """
        Cluster individual feature.

        :param df: DataFrame with original features.
        :param feature_name: String indicating the feature name.
        :param nb_clusters: Integer indicating how many clusters shall be found.
        :return: Numpy array with cluster ids
        """
        self.scalers[feature_name] = MinMaxScaler()
        feat_scaled = self.scalers[feature_name].fit_transform(df[[feature_name]])

        self.cluster_classes[feature_name] = KMeans(
            random_state=self.random_state, n_clusters=nb_clusters
        )
        clusters_found = self.cluster_classes[feature_name].fit_predict(feat_scaled)

        # sort and reindex cluster such as the highest value rewards most points
        temp_df = pd.DataFrame(
            {feature_name: df[feature_name], "cluster": clusters_found}
        )
        reindexed_clusters = self._fit_reindex_clusters_by_mean(
            temp_df, feature_name, higher_is_better
        )

        # move away from 0 index for scoring
        return reindexed_clusters + 1

    def _predict_cluster_feature(
        self, df: pd.DataFrame, feature_name: str
    ) -> np.ndarray:
        """
        Cluster individual feature.

        :param df: DataFrame with original features.
        :param feature_name: String indicating the feature name.
        :return: Numpy array with cluster ids
        """
        feat_scaled = self.scalers[feature_name].transform(df[[feature_name]])

        clusters_found = self.cluster_classes[feature_name].predict(feat_scaled)

        # sort and reindex cluster such as the highest value rewards most points
        temp_df = pd.DataFrame(
            {feature_name: df[feature_name], "cluster": clusters_found}
        )
        reindexed_clusters = self._predict_reindex_clusters_by_mean(
            temp_df, feature_name
        )

        # move away from 0 index for scoring
        return reindexed_clusters + 1

    def fit_predict_cluster(
        self, df: pd.DataFrame, keep_original_features: bool = True
    ):
        """
        Calculate cluster (i.e. RFM) scores based on input features.

        :param df: Pandas DataFrame including the original features. Additional feature will be ignored.
        :param keep_original_features: If true, return clusters and original dataframe. Otherwise return RFM results
            only.
        :return: Pandas DataFrame with RFM scores
        """
        if keep_original_features:
            cluster_results_df = df.copy()
            cluster_results_df.columns = [
                f"{col}_original" for col in cluster_results_df.columns.to_list()
            ]
        else:
            cluster_results_df = pd.DataFrame()

        for feature, cluster_setting in self.cluster_settings.items():
            clusters_found = self._fit_cluster_feature(
                df,
                feature,
                cluster_setting.nb_clusters,
                cluster_setting.higher_is_better,
            )
            cluster_results_df[feature] = clusters_found

        cluster_results_df["total_score"] = cluster_results_df[
            self.cluster_settings.keys()
        ].sum(axis=1)
        return cluster_results_df

    def predict_cluster(self, df: pd.DataFrame, keep_original_features: bool = True):
        """
        Calculate cluster (i.e. RFM) scores based on input features.

        :param df: Pandas DataFrame including the original features. Additional feature will be ignored.
        :param keep_original_features: If true, return clusters and original dataframe. Otherwise return RFM results
            only.
        :return: Pandas DataFrame with RFM scores
        """
        if keep_original_features:
            cluster_results_df = df.copy()
            cluster_results_df.columns = [
                f"{col}_original" for col in cluster_results_df.columns.to_list()
            ]
        else:
            cluster_results_df = pd.DataFrame()

        for feature, _cluster_setting in self.cluster_settings.items():
            clusters_found = self._predict_cluster_feature(df, feature)
            cluster_results_df[feature] = clusters_found

        cluster_results_df["total_score"] = cluster_results_df[
            self.cluster_settings.keys()
        ].sum(axis=1)
        return cluster_results_df



import pandas as pd
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import train_test_split

from bluecast.blueprints.cast import BlueCast
from bluecast.blueprints.cast_regression import BlueCastRegression
from bluecast.config.training_config import TrainingConfig
from bluecast.conformal_prediction.effectiveness_nonconformity_measures import (
    prediction_interval_spans,
)
from bluecast.conformal_prediction.evaluation import (
    prediction_interval_coverage,
    prediction_set_coverage,
)


def test_prediction_set_coverage():
    X, y = make_classification(
        n_samples=10000, n_features=5, random_state=42, n_classes=2
    )
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    X_test, X_calibrate, y_test, y_calibrate = train_test_split(
        X_calibrate, y_calibrate, test_size=0.5, random_state=42
    )

    X_train = pd.DataFrame(X_train)
    X_test = pd.DataFrame(X_test)
    X_calibrate = pd.DataFrame(X_calibrate)

    X_train.columns = X_train.columns.astype(str)
    X_test.columns = X_test.columns.astype(str)
    X_calibrate.columns = X_calibrate.columns.astype(str)

    # Create a custom training config and adjust general training parameters
    train_config = TrainingConfig()
    train_config.hyperparameter_tuning_rounds = 10
    train_config.autotune_model = (
        False  # we want to run just normal training, no hyperparameter tuning
    )

    automl = BlueCast(
        class_problem="binary",
        conf_training=train_config,
    )

    X_train["target"] = y_train
    automl.fit(X_train, target_col="target")

    # make use of calibration
    automl.calibrate(X_calibrate, y_calibrate)

    # prediction sets given a certain confidence interval alpha
    alpha = 0.05
    pred_sets = automl.predict_sets(X_test, alpha=alpha)

    assert prediction_set_coverage(y_test, pred_sets) > 1 - alpha
    assert prediction_set_coverage(pd.Series(y_test), pred_sets) > 1 - alpha


def test_prediction_interval_coverage():
    X, y = make_regression(
        n_samples=10000,
        n_features=5,
        random_state=42,
    )
    X_train, X_calibrate, y_train, y_calibrate = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    X_test, X_calibrate, y_test, y_calibrate = train_test_split(
        X_calibrate, y_calibrate, test_size=0.5, random_state=42
    )
    X_train = pd.DataFrame(X_train)
    X_test = pd.DataFrame(X_test)
    X_calibrate = pd.DataFrame(X_calibrate)

    X_train.columns = X_train.columns.astype(str)
    X_test.columns = X_test.columns.astype(str)
    X_calibrate.columns = X_calibrate.columns.astype(str)

    # Create a custom training config and adjust general training parameters
    train_config_reg = TrainingConfig()
    train_config_reg.hyperparameter_tuning_rounds = 10
    train_config_reg.autotune_model = (
        True  # we want to run just normal training, no hyperparameter tuning
    )

    automl_reg = BlueCastRegression(
        class_problem="regression",
        conf_training=train_config_reg,
    )

    X_train["target"] = y_train
    automl_reg.fit(X_train, target_col="target")

    # make use of calibration
    automl_reg.calibrate(X_calibrate, y_calibrate)

    # p-values for each class being the correct one
    pred_intervals = automl_reg.predict_interval(X_test, alphas=[0.01, 0.05, 0.1])

    val_results = prediction_interval_coverage(
        y_test, pred_intervals, alphas=[0.01, 0.05, 0.1]
    )
    assert val_results[0.01] > val_results[0.05] > val_results[0.1]

    band_sizes = prediction_interval_spans(pred_intervals, alphas=[0.01, 0.05, 0.1])
    assert band_sizes[0.01] > band_sizes[0.05] > band_sizes[0.1]

    val_results = prediction_interval_coverage(
        pd.Series(y_test), pred_intervals, alphas=[0.01, 0.05, 0.1]
    )
    assert val_results[0.01] > val_results[0.05] > val_results[0.1]



from typing import Any, Callable, List

import numpy as np
import pandas as pd

from bluecast.conformal_prediction.base_classes import (
    ConformalPredictionWrapperBaseClass,
)
from bluecast.conformal_prediction.nonconformity_measures_regression import (
    absolute_error,
)


class ConformalPredictionRegressionWrapper(ConformalPredictionWrapperBaseClass):
    """
    Calibrate a model instance given a calibration set.

    :param model: An already fitted model instance of any type
    :param nonconformity_measure_scorer: A function object to calculate nonconformity scores with args
        y_calibration, preds
    """

    def __init__(
        self, model: Any, nonconformity_measure_scorer: Callable = absolute_error
    ):
        self.model = model
        self.nonconformity_measure_scorer = nonconformity_measure_scorer
        self.nonconformity_scores: np.ndarray = np.empty((0, 0))
        self.quantiles: List[float] = []

    def calibrate(self, x_calibration, y_calibration):
        """
        Calibrate a model instance given a calibration set.

        :param x_calibration: Calibration set features. Must be unseen data for the model
        :param y_calibration: Calibration set labels or values
        """
        preds = self.model.predict(x_calibration)
        self.nonconformity_scores = self.nonconformity_measure_scorer(
            y_calibration, preds
        )
        return self.nonconformity_scores

    def predict(self, x):
        return self.model.predict(x)

    def _calculate_intervals(
        self, y_hat: np.ndarray, quantiles: List[float], alphas: List[float]
    ) -> pd.DataFrame:
        """
        Add lower and upper prediction bands for every quantile in quantiles.
        """

        prediction_bands = np.zeros((len(y_hat), 2, len(quantiles)))

        lower_band_cols = []
        higher_band_cols = []
        for i, q in enumerate(quantiles):
            prediction_bands[:, :, i] = np.stack([y_hat - q, y_hat + q], axis=1)
            lower_band_cols.append(f"{alphas[i]}_low")
            higher_band_cols.append(f"{1-alphas[i]}_high")

        lower_preds = pd.DataFrame(prediction_bands[:, 0, :], columns=lower_band_cols)
        upper_preds = pd.DataFrame(prediction_bands[:, 1, :], columns=higher_band_cols)
        all_preds = pd.concat(
            [
                lower_preds,
                upper_preds.reindex(upper_preds.columns.to_list()[::-1], axis=1),
            ],
            axis=1,
        )
        return all_preds

    def predict_interval(self, x: pd.DataFrame, alphas: List[float]) -> pd.DataFrame:
        preds = self.model.predict(x)
        self.quantiles = [
            np.nanquantile(self.nonconformity_scores, 1.0 - alpha, method="higher")
            for alpha in alphas
        ]

        prediction_bands = self._calculate_intervals(preds, self.quantiles, alphas)
        return prediction_bands



from typing import Tuple

import numpy as np
import pandas as pd

from bluecast.conformal_prediction.nonconformity_measures_regression import (
    absolute_error,
    normalized_error,
)


def create_synthetic_regression_data() -> Tuple[np.ndarray, pd.Series]:
    y_true_regression = pd.Series([15.0, 25.0, 40.0])
    synthetic_results_regression = np.asarray([5.0, 25.0, 45.0])
    return synthetic_results_regression, y_true_regression


def test_absolute_error_loss():
    synthetic_results_regression, y_true_regression = create_synthetic_regression_data()
    assert (
        absolute_error(y_true_regression, synthetic_results_regression)
        == np.asarray([10.0, 0.0, 5.0])
    ).all()


def test_normalized_error_error_loss():
    synthetic_results_regression, y_true_regression = create_synthetic_regression_data()
    assert (
        normalized_error(y_true_regression, synthetic_results_regression, 5)
        == np.asarray([2.0, 0.0, 1.0])
    ).all()



from abc import ABC, abstractmethod
from typing import Union

import numpy as np
import pandas as pd


class ConformalPredictionWrapperBaseClass(ABC):
    """Base class for the experiment tracker.

    Enforces the implementation of the add_results and retrieve_results_as_df methods.
    """

    @abstractmethod
    def calibrate(
        self,
        x_calibration: pd.DataFrame,
        y_calibration: Union[pd.Series, np.ndarray],
    ) -> None:
        """
        Calibrate a model instance given a calibration set.

        :param x_calibration: Calibration set features. Must be unseen data for the model
        :param y_calibration: Calibration set labels or values
        """
        pass
